#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
媒体库安装脚本
Create: 2021-01-28
"""
import os
import re
import getpass
import subprocess
import hashlib
import shutil
import json
import platform
from multiprocessing import cpu_count

SCRIPT_PATH = os.path.split(os.path.realpath(__file__))[0]
LIB_SHORT_NAME = ['gmmlib', 'libva', 'media-driver', 'media-sdk']
# 媒体库简称与下载网址对应关系的字典，读取json文件之后，会把内容导入其中。
LIB_WEBSITE = {}
# 媒体库简称与安装包sha256码对应关系的字典，读取json文件之后，会把内容导入其中。
LIB_SHA256_NUM = {}
SHORT_NAME_TO_LONG_NAME = {'gmmlib': ['gmmlib-intel-gmmlib-22.1.2',
                                      'intel-gmmlib-22.1.2'],
                           'libva': ['libva-2.14.0', '2.14.0'],
                           'media-driver': ['media-driver-intel-media-22.3.1',
                                            'intel-media-22.3.1'],
                           'media-sdk': ['MediaSDK-intel-mediasdk-22.3.0',
                                         'intel-mediasdk-22.3.0']}
LIB_EXISTED_BAG = {}


def echo_warning(info):
    """
    功能描述:Warn信息输出格式化
    参数:info 需要输出的信息
    """
    print('\033[33m[ WARN] %s \033[0m' % info)


def echo_info(info):
    """
    功能描述:Info信息输出格式化
    参数:info 需要输出的信息
    """
    print('\033[32m[ INFO] %s \033[0m' % info)


def echo_error(info):
    """
    功能描述:Error信息输出格式化
    参数:info 需要输出的信息
    """
    print('\033[31m[ ERROR] %s \033[0m' % info)


def version_from_string_to_num(version):
    """
    功能描述：把version从string转化为可比较的数字类型
    参数：string类型的版本号
    返回值：数字类型的version
    """
    version_group = version.split('.')
    version_sum = 0
    for version_str in version_group:
        version_sum = version_sum * 100 + int(version_str)
    return version_sum


def test_kernel_version(version_limit):
    """
    功能描述：测试当前系统kernel是否大于或等于version_limit
    参数：版本号界限
    返回值：是，则返回0，否，则返回1
    """
    version = platform.release().split('-')[0]
    version_num = version_from_string_to_num(version)
    version_limit_num = version_from_string_to_num(version_limit)
    if version_num >= version_limit_num:
        return 0
    echo_error("The computer's Linux kernel version is lower than %s. "
               "Please upgrade it to %s or a higher version."
               % (version_limit, version_limit))
    return 1


def process_not_intel_core_cpu():
    """
    功能描述：处理CPU不是Intel core的情况
    """
    cat_cpu_info_process = subprocess.Popen(['cat', '/proc/cpuinfo'],
                                            stdout=subprocess.PIPE,
                                            shell=False)
    grep_name_process = subprocess.Popen(['grep', 'name'],
                                         stdin=cat_cpu_info_process.stdout,
                                         stdout=subprocess.PIPE,
                                         shell=False)
    grep_intel_process = subprocess.Popen(['grep', 'Intel'],
                                          stdin=grep_name_process.stdout,
                                          stdout=subprocess.PIPE,
                                          shell=False)
    cpu_info = grep_intel_process.stdout.readline().decode('utf-8')
    if len(cpu_info) == 0:
        echo_error("The current PC processor is not Intel.")
        return 1
    cpu_info_group = cpu_info.split(":")
    cpu_model_name = cpu_info_group[1]
    echo_warning("The model name and kernel version of "
                 "the current PC are %s and %s, respectively. "
                 "You can query the core name at "
                 "https://www.intel.cn/content/www/cn/zh/homepage.html "
                 "based on the model name. Check whether the CPU is in "
                 "the table \033[3mCPUs That Support Intel "
                 "Hardware Decoding\033[0m \033[33mof "
                 "the\033[0m \033[3;33mMDC Application Visualizer "
                 "User Guide\033[0m \033[33mand "
                 "whether the kernel version satisfies the limit. "
                 "If the CPU is not in the table or "
                 "the kernel version does not satisfy the limit, "
                 "the Intel hardware decoding function cannot be used. "
                 "Then, select Y to continue the "
                 "installation or select N to exit the installation."
                 % (cpu_model_name, platform.release().split('-')[0]))
    choice = "o"
    while len(choice) != 0 and choice not in ('Y', 'N', 'y', 'n'):
        choice = input("[N/y] ")
    if choice in ('Y', 'y'):
        return 0
    return 1


def check_if_cpu_and_kernel_version_suitable():
    """
    功能描述：检查cpu以及内核版本是否符合要求
    """
    cat_cpu_info_process = subprocess.Popen(['cat', '/proc/cpuinfo'],
                                            stdout=subprocess.PIPE,
                                            shell=False)
    grep_name_process = subprocess.Popen(['grep', 'name'],
                                         stdin=cat_cpu_info_process.stdout,
                                         stdout=subprocess.PIPE,
                                         shell=False)
    grep_core_process = subprocess.Popen(['grep', 'Core'],
                                         stdin=grep_name_process.stdout,
                                         stdout=subprocess.PIPE,
                                         shell=False)
    cpu_info = grep_core_process.stdout.readline().decode('utf-8')
    if len(cpu_info) == 0:
        return process_not_intel_core_cpu()
    cpu_info_group = cpu_info.split(" ")
    core_type = None
    for word in cpu_info_group:
        info = re.findall(r'i\d+-[a-zA-Z0-9]*', word)
        if info:
            core_type = info[0]
            break
    core_type_group = core_type.split("-")
    if len(core_type_group) < 2:
        echo_error("The Intel core model is outdated. "
                   "The Intel hardware decoder supports "
                   "only the 6th to 11th Gen Intel cores.")
        return 1
    core_gen = core_type_group[1]
    gen_num = 0
    gen_num = int(core_gen[0])
    result = 1
    if gen_num == 1:
        gen_num = int(core_gen[:2])
    # 参考https://github.com/Intel-Media-SDK/MediaSDK/wiki/Intel-Graphics-Support-in-Linux-Kernels
    # (此网站为github上media-sdk给出Intel CPU代号的内核版本限制表格的网站。）
    if gen_num < 6:
        echo_error("The Intel core model is outdated. "
                   "The Intel hardware decoder supports "
                   "only the 6th to 11th Gen Intel cores.")
    elif gen_num < 8:
        result = test_kernel_version("4.14.0")
    elif gen_num < 10:
        result = test_kernel_version("4.15.0")
    elif gen_num == 10:
        result = test_kernel_version("5.2.0")
    #Ubuntu18.04的内核版本默认是5.4, 5.4不正式支持Intel酷睿11代及以上的产品。
    #Ubuntu20 desktop版本默认内核版本是5.13，对于Intel酷睿11代的机器建议安装Ubuntu20 desktop版本的系统
    elif gen_num == 11:
        result = test_kernel_version("5.5.0")
    elif gen_num > 11:
        echo_error("The Intel hardware decoder does not support "
                   "Intel cores higher than the 11th generation.")
    return result


def check_if_intel_vga_available():
    """
    功能描述：检查当前电脑是否有可用的intel核显
    """
    lspci_process = subprocess.Popen(['lspci'],
                                     stdout=subprocess.PIPE, shell=False)
    grep_vga_process = subprocess.Popen(['grep', 'VGA'],
                                        stdin=lspci_process.stdout,
                                        stdout=subprocess.PIPE, shell=False)
    grep_intel_process = subprocess.Popen(['grep', 'Intel'],
                                          stdin=grep_vga_process.stdout,
                                          stdout=subprocess.PIPE, shell=False)
    result_code = grep_intel_process.wait()
    if result_code != 0:
        echo_error("The computer does not have an Intel graphics "
                   "card or its graphics card is not functional.")
        return 1
    return 0


def check_integrity_of_pkg(lib_short_name):
    """
    功能描述：校验媒体包的完整性
    参数：lib_short_name 要校验的媒体包简称
    """
    sha256_object_of_pkg = hashlib.sha256()
    with open("%s.tar.gz" % LIB_EXISTED_BAG[lib_short_name], 'rb') as pkg:
        for line in pkg:
            sha256_object_of_pkg.update(line)
    sha256_num_of_pkg = sha256_object_of_pkg.hexdigest()
    if sha256_num_of_pkg == LIB_SHA256_NUM[lib_short_name]:
        return 0
    return 1


def modify_media_driver_cmake_flag(media_driver_path):
    """
    功能描述：把media-driver的一个编译设置"-Werror=implicit-function-declaration"删去,
            不删去容易导致在ubuntu系统里安装media-driver失败，而忽略这个警告并不影响media-driver的使用。
    参数：media_driver_path media_driver文件夹目录
    """
    new_file_data = ""
    delete_str = "-Werror=implicit-function-declaration"
    path = os.path.join(SCRIPT_PATH, media_driver_path,
                        'media_driver', 'cmake', 'linux',
                        'media_compile_flags_linux.cmake')
    with open(path, "r") as compile_flags_file:
        all_lines = compile_flags_file.readlines()
        for line in all_lines:
            if delete_str in line:
                continue
            new_file_data += line
    with open(path, "w") as compile_flags_file:
        compile_flags_file.write(new_file_data)


def check_softwares_if_installed():
    """
    功能描述：检查cmake、g++、pkg-config、libpciaccess-dev这四个软件是否已经安装
    """
    result = subprocess.run(['which', 'cmake'], check=False,
                            stdout=subprocess.PIPE)
    if result.returncode != 0:
        echo_error("The CMake software is not installed yet. "
                   "Install CMake 3.6 or a higher version "
                   "and then run the script.")
        return 1
    result = subprocess.run(['which', 'g++'], check=False,
                            stdout=subprocess.PIPE)
    if result.returncode != 0:
        echo_error("The g++ software is not installed yet. "
                   "Run the sudo apt install g++ command "
                   "to install it and then run the script.")
        return 1
    result = subprocess.run(['which', 'pkg-config'], check=False,
                            stdout=subprocess.PIPE)
    if result.returncode != 0:
        echo_error("The pkg-config software is not installed yet. "
                   "Run the sudo apt install pkg-config "
                   "command to install it, and then run the script.")
        return 1
    first_process = subprocess.Popen(['dpkg', '-l'],
                                     stdout=subprocess.PIPE, shell=False)
    second_process = subprocess.Popen(['grep', '-w', 'libpciaccess-dev'],
                                      stdin=first_process.stdout,
                                      stdout=subprocess.PIPE, shell=False)
    result_code = second_process.wait()
    if result_code != 0:
        echo_error("The libpciaccess-dev software is not installed yet. "
                   "Run the sudo apt install libpciaccess-dev "
                   "command to install it, and then run the script.")
        return 1
    return 0


def prepare_for_install():
    """
    功能描述:安装前的环境检查以及设置
    """
    result = check_if_intel_vga_available()
    if result != 0:
        return 1
    result = check_if_cpu_and_kernel_version_suitable()
    if result != 0:
        return 1
    result = check_softwares_if_installed()
    if result != 0:
        return 1
    result = load_json()
    if result != 0:
        return 1
    pkg_config_path = os.getenv('PKG_CONFIG_PATH')
    opt_media_pkgconfig_path = "/opt/media/lib/pkgconfig"
    os.environ['PKG_CONFIG_PATH'] = '%s:%s' % (opt_media_pkgconfig_path,
                                               pkg_config_path)
    return 0


def check_cmd_exists(cmd_name):
    result = os.path.exists(cmd_name)
    if not result:
        install_name = cmd_name.split('/')[-1]
        echo_error(("An error occurred when find '{0}'. "
                   "Please use command 'sudo apt-get install {1}' to install.")
                   .format(cmd_name, install_name))
        return False
    return True


def install(lib_short_name):
    """
    功能描述：安装媒体包
    参数：lib_short_name 媒体包简称
    """
    install_path = os.path.abspath(LIB_EXISTED_BAG[lib_short_name])
    os.mkdir(install_path)
    if check_cmd_exists('/bin/tar') is False:
        return 1
    subprocess.run(['/bin/tar', '-zxf', '%s.tar.gz' %
                   install_path,
                   '--strip-components', '1', '-C',
                   install_path])
    if lib_short_name == "media-driver":
        modify_media_driver_cmake_flag(install_path)
    lib_path = os.path.join(SCRIPT_PATH, install_path)
    work_path = lib_path
    if lib_short_name == 'libva':
        result = subprocess.run(['./autogen.sh',
                                '--prefix=/opt/media'],
                                cwd=lib_path)
        if result.returncode != 0:
            echo_error("An error occurred when libva runs autogen.sh.")
            return 1
    else:
        os.makedirs(os.path.join(install_path, 'build'), exist_ok=True)
        work_path = os.path.join(SCRIPT_PATH, install_path, 'build')
        if check_cmd_exists('/usr/bin/cmake') is False:
            return 1
        result = subprocess.run(['/usr/bin/cmake',
                                lib_path,
                                '-DCMAKE_INSTALL_PREFIX=/opt/media'],
                                cwd=work_path)
        if result.returncode != 0:
            echo_error("An error occurred when %s "
                       "runs the cmake command." % lib_short_name)
            return 1
    if check_cmd_exists('/usr/bin/make') is False:
        return 1
    result = subprocess.run(['/usr/bin/make', '-j%d' % cpu_count()], cwd=work_path)
    if result.returncode != 0:
        echo_error("An error occurred when %s "
                   "runs the make -j%d command." %
                   (lib_short_name, cpu_count()))
        return 1
    result = subprocess.run(['/usr/bin/make', 'install'], cwd=work_path)
    if result.returncode != 0:
        echo_error("An error occurred when %s "
                   "runs the make install command." % lib_short_name)
        return 1
    return 0


def check_pakage_exist_and_integrity(lib_short_name):
    """
    功能描述：检查media_src文件夹下媒体包是否存在以及校验它的完整性
    参数：lib_short_name 媒体包的简称
    """
    bag_existed = False
    for lib_long_name in SHORT_NAME_TO_LONG_NAME[lib_short_name]:
        if os.path.exists(os.path.join(SCRIPT_PATH, '%s.tar.gz' %
                                                    lib_long_name)):
            LIB_EXISTED_BAG.update({lib_short_name: lib_long_name})
            bag_existed = True
            break
    if not bag_existed:
        bag_name = '%s.tar.gz' % SHORT_NAME_TO_LONG_NAME[lib_short_name][0]
        echo_error("No local package %s is available. "
                   "Download the complete %s from %s, and copy it to the "
                   "%s "
                   "directory." %
                   (bag_name, bag_name, LIB_WEBSITE[lib_short_name],
                    SCRIPT_PATH))
        return 1
    result = check_integrity_of_pkg(lib_short_name)
    if result != 0:
        bag_name = '%s.tar.gz' % LIB_EXISTED_BAG[lib_short_name]
        echo_error("The local package %s is incomplete. "
                   "Download the complete %s from %s, and copy it to the "
                   "%s "
                   "directory." %
                   (bag_name, bag_name, LIB_WEBSITE[lib_short_name],
                    SCRIPT_PATH))
        return 1
    return 0


def load_jason_for_each_lib(jason_map, lib_short_name):
    """
    功能描述：从配置文件解析后的对象导入某个媒体库的下载网址和sha256码
    参数：jason_map 配置文件解析后的对象
         lib_short_name 媒体包简称
    """
    lib_temp = jason_map.get(lib_short_name, 'none')
    if lib_temp == 'none':
        echo_error("The lib_information.json file is modified. "
                   "%s cannot be found." % lib_short_name)
        return 1
    lib_website = lib_temp.get('url', 'none')
    if lib_website == 'none':
        echo_error("The lib_information.json file is modified. "
                   "The url component of %s cannot be found." % lib_short_name)
        return 1
    lib_sha256_number = lib_temp.get('sha256', 'none')
    if lib_sha256_number == 'none':
        echo_error("The lib_information.json file is modified. "
                   "The sha256 component of %s cannot be found." %
                   lib_short_name)
        return 1
    LIB_WEBSITE.update({lib_short_name: lib_website})
    LIB_SHA256_NUM.update({lib_short_name: lib_sha256_number})
    return 0


def load_json():
    """
    功能描述：从配置文件lib_information.json中导入媒体包的下载网址和sha256码
    """
    if not os.path.exists(os.path.join(SCRIPT_PATH, 'lib_information.json')):
        echo_error("The lib_information.json file does not exist.")
        return 1
    with open("lib_information.json", 'r') as lib_json:
        json_content = lib_json.read()
    try:
        temp = json.loads(json_content)
    except ValueError:
        echo_error("The lib_information.json file is modified. "
                   "A parsing error occurred.")
        return 1
    for lib_short_name in LIB_SHORT_NAME:
        result = load_jason_for_each_lib(temp, lib_short_name)
        if result != 0:
            return 1
    return 0


def install_lib(lib_short_name):
    """
    功能描述：完成对某个媒体包的整个安装过程
    参数：lib_short_name 媒体包简称
    """
    result = check_pakage_exist_and_integrity(lib_short_name)
    if result != 0:
        return 1
    shutil.rmtree(LIB_EXISTED_BAG[lib_short_name], ignore_errors=True)
    result = install(lib_short_name)
    if result != 0:
        return 1
    shutil.rmtree(LIB_EXISTED_BAG[lib_short_name], ignore_errors=True)
    return 0


def main():
    """
    功能描述：媒体库安装脚本的入口函数
    """
    if getpass.getuser() != 'root':
        print("Use the sudo ./install_intel_media_stack.py "
              "command to run the script.")
        return
    result = prepare_for_install()
    if result != 0:
        return
    for lib_short_name in LIB_SHORT_NAME:
        result = install_lib(lib_short_name)
        if result != 0:
            return
    echo_info("Installed successfully. ")


if __name__ == "__main__":
    main()
