/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2022. All rights reserved.
 * Description: 用于向MViz所在的上位机发布和订阅自定义数据类型
 * Create: 2020-11-30
 */
#ifndef MVIZ_DEVEL_KIT_H
#define MVIZ_DEVEL_KIT_H

#include <memory>
#include <boost/bind/bind.hpp>
#include <ros/subscribe_options.h>
#include <securec.h>

#include <publisher.h>
#include <subscriber.h>

#include "logger.h"

namespace mviz {
const int32_t MAX_MALLOC_SIZE = 80 * 1024 * 1024;

/*
 * 自定义数据发布和订阅时的消息特征
 */
class MessageTraits {
public:
    MessageTraits() = default;
    ~MessageTraits() = default;

    /*
     * 功能：设置md5sum的值
     * 输入参数：value-将要设置的值
     * 输出参数：无
     * 返回值：无
     */
    void SetMd5Sum(const std::string &value) noexcept
    {
        md5sum = value;
    }

    /*
     * 功能：设置datatype的值
     * 输入参数：value-将要设置的值
     * 输出参数：无
     * 返回值：无
     */
    void SetDatatype(const std::string &value) noexcept
    {
        datatype = value;
    }

    /*
     * 功能：设置messageDefinition的值
     * 输入参数：value-将要设置的值
     * 输出参数：无
     * 返回值：无
     */
    void SetMessageDefinition(const std::string &value) noexcept
    {
        messageDefinition = value;
    }

    /*
     * 功能：设置hasHeader的值
     * 输入参数：value-将要设置的值
     * 输出参数：无
     * 返回值：无
     */
    void SetHasHeader(bool value) noexcept
    {
        hasHeader = value;
    }

    /*
     * 功能：将MessageTraits序列化为二进制数据
     * 输入参数：无
     * 输出参数：无
     * 返回值：serializedTraits-序列化后得到的二进制数据
     */
    std::vector<uint8_t> SerializeMessageTraits() const noexcept
    {
        using namespace ros::serialization;

        auto serializeLength = Serializer<std::string>::serializedLength(md5sum);
        serializeLength += Serializer<std::string>::serializedLength(datatype);
        serializeLength += Serializer<std::string>::serializedLength(messageDefinition);
        serializeLength += Serializer<bool>::serializedLength(hasHeader);

        std::vector<uint8_t> serializedTraits;
        serializedTraits.resize(serializeLength);
        OStream oStream(&serializedTraits[0], serializeLength);
        Serializer<std::string>::write(oStream, md5sum);
        Serializer<std::string>::write(oStream, datatype);
        Serializer<std::string>::write(oStream, messageDefinition);
        Serializer<bool>::write(oStream, hasHeader);
        return serializedTraits;
    }
private:
    std::string md5sum {};
    std::string datatype {};
    std::string messageDefinition {};
    bool hasHeader { false };
};

class Subscriber {
public:
    Subscriber() = default;
    ~Subscriber() = default;

    /*
     * 功能：初始化Subscriber，构造消息特征的二进制数据，并将其发送给MViz服务端
     * 输入参数：options-包含用于订阅消息的topic/md5sum/datatype等参数
     * 输出参数：无
     * 返回值：无
     */
    template <class M>
    void Init(const ros::SubscribeOptions &options) noexcept
    {
        MessageTraits traits {};
        traits.SetMd5Sum(options.md5sum);
        traits.SetDatatype(options.datatype);
        traits.SetHasHeader(ros::message_traits::hasHeader<M>());

        std::vector<uint8_t> serializedMessageTraits = traits.SerializeMessageTraits();

        auto callback = std::bind(&Subscriber::InternalCallback, options.topic, options.helper,
            std::placeholders::_1, std::placeholders::_2);
        subscriber = std::make_shared<mdc::visual::Subscriber>(options.topic, &serializedMessageTraits[0],
            serializedMessageTraits.size(), callback);
    }

private:
    std::shared_ptr<mdc::visual::Subscriber> subscriber { nullptr };

    static void InternalCallback(const std::string &topic, ros::SubscriptionCallbackHelperPtr helper,
        const uint8_t buffer[], uint32_t len) noexcept
    {
        ros::SerializedMessage serializedMessage {};
        serializedMessage.num_bytes = len;
        if (len <= 0 || len > MAX_MALLOC_SIZE) {
            Logger::Error("Memory allocation failed! The num_bytes of message is an invalid value, "
                "the value of serializedMessage.num_bytes is " + std::to_string(len) + ", "
                "value should be in the range of 1 to " + std::to_string(MAX_MALLOC_SIZE));
            return;
        }
        auto buf = new uint8_t[len];
        serializedMessage.buf = boost::shared_array<uint8_t>(buf);

        auto ret = memcpy_s(serializedMessage.buf.get(), serializedMessage.num_bytes, buffer, len);
        if (ret != EOK) {
            Logger::Error("Failed to copy the memory. Error code: " +  std::to_string(ret) +
                ". Memory may be insufficient. Release memory and restart this application.");
            return;
        }

        ros::VoidConstPtr msg;
        try {
            ros::SubscriptionCallbackHelperDeserializeParams params;
            params.buffer = serializedMessage.buf.get();
            params.length = serializedMessage.num_bytes;
            msg = helper->deserialize(params);
        } catch (std::exception& e) {
            Logger::Error("An exception occurred during the deserialization of the data whose topic is " +
                topic + ". Exception: " + e.what() + ".");
            return;
        }

        ros::SubscriptionCallbackHelperCallParams params;
        params.event = ros::MessageEvent<void const>(msg,
            boost::shared_ptr<ros::M_string>(),
            ros::Time(),
            true,
            ros::MessageEvent<void const>::CreateFunction());
        helper->call(params);
    }
};

class Publisher {
public:
    Publisher() = default;
    ~Publisher() = default;

    /*
     * 功能：初始化Publisher，构造消息特征的二进制数据
     * 输入参数：topic-用于发布消息的Topic
     * 输出参数：无
     * 返回值：无
     */
    template <class M>
    void Init(const std::string &topic) noexcept
    {
        MessageTraits traits {};
        traits.SetMd5Sum(ros::message_traits::md5sum<M>());
        traits.SetDatatype(ros::message_traits::datatype<M>());
        traits.SetMessageDefinition(ros::message_traits::definition<M>());
        traits.SetHasHeader(ros::message_traits::hasHeader<M>());

        serializedTraits = traits.SerializeMessageTraits();

        publisher = mdc::visual::Publisher::Advertise<uint8_t const *>(topic);
    }

    /*
     * 功能：发布用户数据到上位机的ROS中
     * 输入参数：message-用户创建的消息实例
     * 输出参数：无
     * 返回值：无
     */
    template <class M>
    void Publish(const M& message) noexcept
    {
        if (serializedTraits.empty()) {
            Logger::Error("Cannot send the message because mviz::Publisher is not initialized.");
            return;
        }
        ros::SerializedMessage serializedMessage = ros::serialization::serializeMessage(message);

        auto dataLength = serializedTraits.size() + serializedMessage.num_bytes;
        auto data = std::make_unique<uint8_t[]>(dataLength);

        auto ret = memcpy_s(data.get(), dataLength, &serializedTraits[0], serializedTraits.size());
        if (ret != EOK) {
            Logger::Error("Failed to copy the memory. Error code: " +  std::to_string(ret) +
                ". Memory may be insufficient. Release memory and restart this application.");
            return;
        }
        ret = memcpy_s(data.get() + serializedTraits.size(), dataLength - serializedTraits.size(),
            serializedMessage.buf.get(), serializedMessage.num_bytes);
        if (ret != EOK) {
            Logger::Error("Failed to copy the memory. Error code: " +  std::to_string(ret) +
                ". Memory may be insufficient. Release memory and restart this application.");
            return;
        }

        publisher.Publish(data.get(), dataLength);
    }

private:
    mdc::visual::Publisher publisher {};
    std::vector<uint8_t> serializedTraits {};
};

class NodeHandle {
public:
    NodeHandle() = default;
    ~NodeHandle() = default;

    /*
     * 功能：封装viz_lib的Connect函数，并提示连接是否成功，和提示MViz Server地址
     * 输入参数：无
     * 输出参数：无
     * 返回值：无
     */
    static void Connect() noexcept
    {
        bool isConnected = mdc::visual::Connect();
        if (isConnected) {
            Logger::Info("The MViz server address is parsed and the visualization client is created successfully.");
        } else {
            Logger::Error("Failed to parse the MViz server address, or a visualization client has been created.");
        }
    }

    /*
     * 功能：封装viz_lib的Close函数，调用mviz::NodeHandle::Connect()后，在退出前应调用mviz::NodeHandle::Close()
     * 输入参数：无
     * 输出参数：无
     * 返回值：无
     */
    static void Close() noexcept
    {
        mdc::visual::Close();
    }

    /*
     * 功能：根据Topic实例化Publisher，用于发送用户自定义数据
     * 输入参数：topic-用户用于发布数据的Topic
     * 输出参数：无
     * 返回值：Publisher-根据Topic实例化得到的Publisher
     */
    template <class M>
    Publisher Advertise(const std::string &topic) noexcept
    {
        Publisher publisher;
        publisher.Init<M>(topic);
        return publisher;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，
     *         callback-用户用于接收数据的回调函数，回调函数为非const的成员函数，入参为非顶层const类型
     *         object-回调函数所属的对象的指针
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M, class T>
    Subscriber Subscribe(const std::string& topic, void(T::*fp)(M), T *object) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.initByFullCallbackType<M>(topic, 0, boost::bind(fp, object, boost::placeholders::_1));
        subscriber.Init<typename ros::ParameterAdapter<M>::Message>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，
     *         callback-用户用于接收数据的回调函数，回调函数为const的成员函数，入参为非顶层const类型
     *         object-回调函数所属的对象的指针
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M, class T>
    Subscriber Subscribe(const std::string& topic, void(T::*fp)(M) const, T *object) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.initByFullCallbackType<M>(topic, 0, boost::bind(fp, object,  boost::placeholders::_1));
        subscriber.Init<typename ros::ParameterAdapter<M>::Message>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，
     *         callback-用户用于接收数据的回调函数，回调函数为非const的成员函数，入参为顶层const类型
     *         object-回调函数所属的对象的指针
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M, class T>
    Subscriber Subscribe(const std::string& topic, void(T::*fp)(const boost::shared_ptr<M const>&), T *object) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.init<M>(topic, 0, boost::bind(fp, object,  boost::placeholders::_1));
        subscriber.Init<M>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，
     *         callback-用户用于接收数据的回调函数，回调函数为const的成员函数，入参为顶层const类型
     *         object-回调函数所属的对象的指针
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M, class T>
    Subscriber Subscribe(const std::string& topic, void(T::*fp)(const boost::shared_ptr<M const>&) const,
        T *object) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.init<M>(topic, 0, boost::bind(fp, object,  boost::placeholders::_1));
        subscriber.Init<M>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，
     *         callback-用户用于接收数据的回调函数，回调函数为boost::function类型，入参为非顶层const类型
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M>
    Subscriber Subscribe(const std::string &topic, const boost::function<void (M)> &callback) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.initByFullCallbackType<M>(topic, 0, callback);
        subscriber.Init<typename ros::ParameterAdapter<M>::Message>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，
     *         callback-用户用于接收数据的回调函数，回调函数为boost::function类型，入参为顶层const类型
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M>
    Subscriber Subscribe(const std::string &topic,
        const boost::function<void(const boost::shared_ptr<M const>&)> &callback) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.init<M>(topic, 0, callback);
        subscriber.Init<M>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，fp-用户用于接收数据的回调函数，回调函数为普通函数类型，入参为非顶层const类型
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M>
    Subscriber Subscribe(const std::string &topic, void(*fp)(M)) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.initByFullCallbackType<M>(topic, 0, fp);
        subscriber.Init<typename ros::ParameterAdapter<M>::Message>(ops);
        return subscriber;
    }

    /*
     * 功能：根据Topic和callback实例化Subscriber，用于接收数据
     * 输入参数：topic-用户用于订阅数据的Topic名，fp-用户用于接收数据的回调函数，回调函数为普通函数类型，入参为顶层const类型
     * 输出参数：无
     * 返回值：Subscriber-根据Topic和回调函数实例化得到的Subscriber
     */
    template<class M>
    Subscriber Subscribe(const std::string &topic, void(*fp)(const boost::shared_ptr<M const>&)) noexcept
    {
        Subscriber subscriber;
        ros::SubscribeOptions ops;
        ops.init<M>(topic, 0, fp);
        subscriber.Init<M>(ops);
        return subscriber;
    }
};
}

#endif // MVIZ_DEVEL_KIT_H
