#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright Huawei Technologies Co., Ltd. 2020-2021. All rights reserved.
"""
import os
import sys
from optparse import OptionParser
import mviz_genmsg


class GenerationProperty:
    """
    用于存放消息生成时将用到的各属性项，包括package_name，include_path等
    """
    def __init__(self):
        self.include_path = ''
        self.template_dir = ''
        self.output_dir = ''
        self.package_name = ''
        self.msgs_file_dir = ''
        self.output_path = ''


def package_dirs_to_list(package_dirs):
    """
    将按空格分割的字符串转为列表
    :param package_dirs: 使用空格分割的多个路径的字符串
    :return: 由路径组成的列表
    """
    return package_dirs.split(' ')


def append_dirs_to_include_path(dirs, properties):
    """
    将输入的文件夹都添加到include_path中,include_path中可能已经包含了值
    :param dirs: 输入的文件夹
    :param properties: 存放include_path的properties
    """
    correct_dirs = []
    for dir_path in dirs:
        if not os.path.exists(dir_path):
            mviz_genmsg.warn('%s does not exist and will be skipped by the '
                             'message code generation tool.' % dir_path)
            continue
        if os.path.isfile(dir_path):
            mviz_genmsg.warn('%s is not a folder and will be skipped by the '
                             'message code generation tool.' % dir_path)
            continue
        msg_path = os.path.join(dir_path, 'msg')
        if not os.path.isdir(msg_path):
            mviz_genmsg.warn('The %s directory does not contain the msg '
                             'folder and will be skipped by the message code '
                             'generation tool.' % dir_path)
            continue

        dir_name = os.path.basename(dir_path)
        properties.include_path.append('%s:%s' % (dir_name, msg_path))
        correct_dirs.append(dir_path)
    return correct_dirs


def execute_message_generation(dirs, properties):
    """
    为每个文件夹执行消息生成
    :param dirs: 输入的文件夹
    :param properties: 存放了所有属性项的properties
    """
    for dir_path in dirs:
        msg_path = os.path.join(dir_path, 'msg')
        dir_name = os.path.basename(dir_path)
        properties.package_name = dir_name
        properties.msgs_file_dir = msg_path
        properties.output_dir = os.path.join(properties.output_path, 'include',
                                             properties.package_name)

        mviz_genmsg.genmsg_by_package(properties)


def main():
    """
    为输入的所有文件夹生成消息代码
    """
    parser = OptionParser("[options]")
    parser.add_option("-d", dest='package_dirs',
                      help="The value of this parameter is a character string "
                           "consisting of the paths of multiple message "
                           "packages. Separate the message packages with "
                           "a space.")
    parser.add_option("-e", dest='template_dir',
                      help="Indicates the folder storing the template file.")
    parser.add_option("-I", dest='include_path',
                      help="Indicates the path included during message "
                           "code generation.",
                      action="append")
    parser.add_option("-o", dest='output_path',
                      help="Indicates the output path of the generated "
                           "message code.")
    (options, argv) = parser.parse_args(sys.argv)

    properties = GenerationProperty()
    properties.include_path = options.include_path
    properties.template_dir = options.template_dir
    properties.output_path = options.output_path
    dirs = package_dirs_to_list(options.package_dirs)

    # 通过将输入的文件夹都添加到include_path中，可允许多个package中的.msg文件相互引用
    correct_dirs = append_dirs_to_include_path(dirs, properties)

    execute_message_generation(correct_dirs, properties)


if __name__ == "__main__":
    main()
