#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright Huawei Technologies Co., Ltd. 2020-2022. All rights reserved.
"""
import glob
import os
import sys
import shutil
from optparse import OptionParser
import genmsg.template_tools


def info(msg):
    """
    打印INFO级别的日志
    :param msg: 日志内容
    """
    print('[Code Generation Tool][INFO] %s' % msg)


def warn(msg):
    """
    打印WARN级别的日志
    :param msg: 日志内容
    """
    print('[Code Generation Tool][WARN] %s' % msg)


def is_same_file(lhs, rhs):
    """
    判断两个文件的内容是否相同
    :param lhs: 被比较的文件路径
    :param rhs: 用于比较的文件路径
    :return: 文件内容是否相同
    """
    with open(lhs, 'rb') as lhs_file:
        with open(rhs, 'rb') as rhs_file:
            return lhs_file.read() == rhs_file.read()


def is_header_exist(directory, msg_file_name):
    """
    检查directory中msg_file_name对应的.h文件是否存在
    :param directory: 检查路径
    :param msg_file_name: .msg文件的文件名
    :return: 对应的.h文件是否存在
    """
    name = msg_file_name.split('.')[0]
    header_file_path = os.path.join(directory, '%s.h' % name)
    return os.path.exists(header_file_path)


def is_any_msg_file_changed(msg_files, backup_dir, output_dir):
    """
    检查是否有.msg文件被修改过，检查需要对比缓存的.msg文件存在且与用户的.msg文件是否一致，
    以及.h文件是否存在
    :param msg_files: 所有.msg文件的绝对路径列表
    :param backup_dir: 构建缓存目录路径，存放.msg文件的一份备份，用于做是否修改过的对比
    :param output_dir: .h文件输出路径
    :return: 存在任何一个被修改过的文件就返回True，否则False
    """
    is_changed = False
    for file_path in msg_files:
        file_name = os.path.basename(file_path)
        backup_file_path = os.path.join(backup_dir, file_name)
        if os.path.exists(backup_file_path) \
                and is_same_file(file_path, backup_file_path) \
                and is_header_exist(output_dir, file_name):
            continue
        is_changed = True
    return is_changed


def backup_msg_file(file_path, backup_dir):
    """
    :param file_path:
    :param backup_dir:
    """
    file_name = os.path.basename(file_path)
    backup_file_path = os.path.join(backup_dir, file_name)
    shutil.copyfile(file_path, backup_file_path)


def genmsg_by_package(ops):
    """
    遍历ops.msgs_file_dir目录中所有的.msg文件，并将其生成为.h文件
    :param ops: 用于生成消息头文件的参数
    """
    if not ops.package_name:
        info('No package_name specified.')
        return

    msg_template_map = {'msg.h.template': '@NAME@.h'}
    srv_template_map = {'srv.h.template': '@NAME@.h'}

    backup_dir = os.path.join('msg_file_backup', ops.package_name)

    if not os.path.exists(backup_dir):
        os.makedirs(backup_dir)

    msg_files = glob.glob(os.path.join(ops.msgs_file_dir, '*.msg'))
    if not is_any_msg_file_changed(msg_files, backup_dir, ops.output_dir):
        return

    for file_path in msg_files:
        backup_msg_file(file_path, backup_dir)

        file_name = os.path.basename(file_path)
        genmsg.template_tools.generate_from_file(
            file_path,
            ops.package_name,
            ops.output_dir,
            ops.template_dir,
            ops.include_path,
            msg_template_map,
            srv_template_map)
        info("Generating C++ code from %s/%s" % (ops.package_name, file_name))


def main():
    """
    为输入的单个文件生成消息代码
    """
    parser = OptionParser("[options]")
    parser.add_option("-m", dest='msgs_file_dir',
                      help="Indicates the message file used to "
                           "generate C++ code.")
    parser.add_option("-p", dest='package_name',
                      help="Indicates the package name of message files.")
    parser.add_option("-o", dest='output_dir',
                      help="Indicates the output path of the generated "
                           "message code.")
    parser.add_option("-e", dest='template_dir',
                      help="Indicates the folder storing the template file.")
    parser.add_option("-I", dest='include_path',
                      help="Indicates the path included during message "
                           "code generation.",
                      action="append")

    (options, argv) = parser.parse_args(sys.argv)
    genmsg_by_package(options)


if __name__ == "__main__":
    main()
