#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright Huawei Technologies Co., Ltd. 2020-2023. All rights reserved.
mviz 启动脚本
"""
import argparse
import json
import os
import platform
import signal
import subprocess
import sys
import time
import resource

ros_distro = ''
required_distro = ['Ubuntu']
python_version_for_ros = ''
ros_env_check_notice = ""

REQUIRED_SYSTEM = 'Linux'
DEFAULT_FLAGS = {'is_block': False, 'is_output_log': False}
SCRIPT_PATH = os.path.split(os.path.realpath(__file__))[0]
PROC_PATH = '/proc'


class RosNoetic:
    """
    功能描述:noetic版本对应的属性值
    """

    def __init__(self):
        self.ros_distro = 'noetic'
        self.python_version_for_ros = 'python3'


def set_check_notice_words():
    """
    功能描述:环境检查打印
    参数:无
    """
    global ros_env_check_notice
    ros_env_check_notice = """Some Dependent dynamic library not found. \
Check that ROS is installed correctly. \n \
If the ROS is not installed, install the ros-""" + \
                           ros_distro + """-desktop-full version. \n \
\n \
If ros-""" + ros_distro + """-desktop has been installed, \
install ros-""" + ros_distro + """-desktop-full and try again.
    (You can type: sudo apt install ros-""" + \
                           ros_distro + """-desktop-full command.) \n \
\n \
If the ROS has been installed, export the environment variables of \
the ROS and try again. \n \
    (You can type: source /opt/ros/""" + \
                           ros_distro + """/setup.bash command.) \n \
\n \
For more information, see MDC Application Visualizer User Guide."""


class BackgroundVar:
    """
    功能描述:新进程、log文件、log窗口记录类
    """

    def __init__(self, process_set):
        """
        功能描述:类初始化
        参数:process_set  进程集合
        """
        self.process_set = process_set

    def clean_background(self):
        """
        功能描述:清除所有由此脚本拉起的进程
        参数:无
        """
        failed_count = 0
        info('Clearing the MViz runtime environment...')
        # 为了避免roscore关闭后，MViz进程中设置ros参数的部分一直尝试连接rosmaster，导致MViz不能正常退出，
        # 所以这里需要对存储进程的list进行反转，这样可以先关闭MViz，再退出roscore
        for single_process in reversed(self.process_set):
            try:
                single_process.send_signal(signal.SIGINT)
                single_process.wait()
            except OSError:
                # 将执行的命令打印出来
                cmd = single_process.args[0] if len(single_process.args) > 0 else ''
                cmd = os.path.split(cmd)[-1]
                warning('Failed to end the process ' + cmd)
                failed_count += 1

        if failed_count == 0:
            info('The MViz runtime environment has been cleared.')
        else:
            warning('The MViz runtime environment has been cleared with %d failed. '
                    'You can end them by command: kill -9.' % failed_count)
        return 0


def waiting(interval, text='Loading'):
    """
    功能描述:加载程序时的动画信息
    参数:interval 动画显示时间, 单位为秒
        text  需要显示的文案
    """
    i = 0
    total_loop_times = interval * 10
    single_loop_times = 4
    while i < total_loop_times:
        for j in '\\' '|' '/' '-':
            sys.stdout.write(("\033[32m[ INFO] %s%s%s%s%s"
                              " %s ... %s%s%s%s%s  "
                              "\033[0m\r" % (j, j, j, j, j,
                                             text, j, j, j, j, j)))
            time.sleep(0.1)
        i = i + single_loop_times


def info(data):
    """
    功能描述:Info信息输出格式化
    参数:data 需要输出的信息
    """
    print('\033[32m[ INFO] %s \033[0m' % data)


def warning(data):
    """
    功能描述:Warn信息输出格式化
    参数:data 需要输出的信息
    """
    print('\033[31m[ WARN] %s \033[0m' % data)


def get_process_name_by_pid(pid):
    """
    功能描述:通过进程的pid获取进程名
    参数:pid 进程的pid
    """
    stat_file = '%s/%s/stat' % (PROC_PATH, pid)
    try:
        with open(stat_file, "rb") as f:
            data = f.read()
    except IOError:
        return
    right_par = data.rfind(b')')
    name = data[data.find(b'(') + 1:right_par]
    return name.decode("utf-8")


def is_process_running(proc_name):
    """
    功能描述:通过进程名，检测进程是否正在运行
    参数:proc_name 进程名
    """
    pid_list = [int(x) for x in os.listdir(PROC_PATH) if x.isdigit()]
    for pid in pid_list:
        name = get_process_name_by_pid(pid)
        if name == proc_name:
            return True
    return False


def launch_cmd(background_var,
               cmd,
               process,
               flags):
    """
    功能描述:执行命令并根据成功与否输出log信息
    参数:background_var 储存运行在后台的进程
        cmd            执行程序的路径和参数
        process    需要执行的命令
        flags          是否阻塞/是否在启动前sleep的标志位字典/是否打印日志到屏幕
    """

    def run_fail_log(process, log=None):
        """
        功能描述:启动程序失败输出的信息
        参数:process 程序名字
            log     程序log记录位置
        """
        if log is None:
            warning('Failed to start the %s process.' % process)
            if process == 'MViz':
                warning('Download the latest MViz installation package to ensure MViz integrity.')
            else:
                warning(ros_env_check_notice)
        else:
            warning('Failed to start the %s process. The startup log of this '
                    'process is stored in the %s file. Check the error '
                    'information in the log file.' % (process, log))

    info('Attempting to start %s...' % process)
    try:
        if flags['is_block']:
            result = subprocess.run(cmd)
            if result.returncode == 1:
                run_fail_log(process)
                return 1

            info('%s started successfully.' % process)
            return 0
        else:
            if flags['is_output_log']:
                single_process = subprocess.Popen(cmd)
            else:
                single_process = subprocess.Popen(
                    cmd, stdout=subprocess.DEVNULL)

            waiting(2, 'Loading ' + process)
            single_process.poll()
            return_code = single_process.returncode
            if return_code is not None and return_code > 0:
                run_fail_log(process)
                return 1
            background_var.process_set.append(single_process)

            real_process_name = os.path.basename(cmd[0])
            while not is_process_running(real_process_name):
                waiting(2, 'Loading ' + process)

            info('%s started successfully.' % process)
            return 0
    except (subprocess.CalledProcessError, IOError) as e:
        warning(e)
        run_fail_log(process)
        return 1
    except FileNotFoundError:
        warning('Failed to find the %s command.' % process)
        run_fail_log(process)
        return 1
    except Exception as e:
        warning(str(e))
        return 1


def query_os_version_info(file):
    """
    功能描述:遍历文件查找系统版本信息
    参数:file 存放系统版本信息的文件
    """

    version_info = {}
    attribute_prefix = 'DISTRIB_'
    prefix_len = len(attribute_prefix)
    for line in file:
        line = line.strip()
        if '=' not in line:
            continue
        key, val = line.split('=', 1)
        if key.startswith(attribute_prefix):
            key = key[prefix_len:]
            if val:
                version_info[key] = val.strip()
    return version_info


def get_os_version():
    """
    功能描述:获取当前系统的类型和版本号
    参数:无
    """

    version_info = {}
    os_info_file = os.environ.get('LSB_ETC_LSB_RELEASE', '/etc/lsb-release')
    if not os.path.exists(os_info_file):
        warning('No such file or directory: ' + os_info_file)
        return version_info
    try:
        with open(os_info_file, 'r') as file:
            version_info = query_os_version_info(file)
    except IOError as msg:
        warning('Unable to open ' + os_info_file + ':', str(msg), file=sys.stderr)

    return version_info


def check_ros_and_ubuntu_version(background_var):
    """
    功能描述:检测用户os版本和ros版本信息是否匹配
    参数:background_var 储存运行在后台的进程
    """

    def check_os_and_package_version() -> bool:
        """
       功能描述:检查当前MViz包对应的运行环境
       参数:当前系统ubuntu版本
       """
        file_path = os.path.join(SCRIPT_PATH, 'version.json')
        version_json = FileHandler(file_path)
        file_content = version_json.get_file_contents()
        if "RunOsVersion" not in file_content:
            warning('The keyword \"RunOsVersion\" cannot be found '
                    'in the %s file.' % file_path)
            return False
        fields_content = file_content.get("RunOsVersion")
        run_os_version = fields_content.split("Ubuntu")
        if len(run_os_version) < 2:
            warning('The value of the \"RunOsVersion\" field (%s) '
                    'in the %s file is incorrect.' %
                    (fields_content, file_path))
            return False
        package_version = run_os_version[1]
        global required_distro
        required_distro.insert(1, package_version)
        global ros_distro
        global python_version_for_ros
        if package_version == "20.04":
            ros_noetic = RosNoetic()
            ros_distro = ros_noetic.ros_distro
            python_version_for_ros = ros_noetic.python_version_for_ros
        else:
            warning('The value of the \"RunOsVersion\" field (%s) '
                    'in the %s file is incorrect.' %
                    (fields_content, file_path))
            return False
        set_check_notice_words()
        return True

    def check_user_os_distro():
        """
       功能描述:检测用户os版本信息
       参数:无
       """
        current_system = platform.system()
        if REQUIRED_SYSTEM != current_system:
            warning('MViz cannot start because the current operating system (%s) '
                    'is not Linux.' % current_system)
            return 1

        os_version = get_os_version()
        current_os = os_version.get('ID')
        if current_os != required_distro[0]:
            warning('MViz cannot start because the current operating system (%s) is not '
                    'Ubuntu.' % current_os)
            return 1

        current_ubuntu = os_version.get('RELEASE')
        if not check_os_and_package_version():
            return 1
        if current_ubuntu != required_distro[1]:
            warning('MViz cannot start because the current version of the '
                    'operating system (%s) is not '
                    '%s.' % (current_ubuntu, required_distro[1]))
            return 1

        info('The current version of the operating system is %s %s.'
             % (required_distro[0], required_distro[1]))
        return 0

    def get_user_ros_distro():
        """
        功能描述:检测用户ros版本信息
        参数:无
        """
        check_ros_cmd = ['rosparam', 'get', 'rosdistro']
        try:
            version_output = subprocess.check_output(
                check_ros_cmd, stderr=subprocess.STDOUT)
            version_output = version_output.replace(
                str.encode('\n'), str.encode(''))
            version_output = version_output.replace(
                str.encode("'"), str.encode(''))
            version_output = version_output.strip()
            ros_version = bytes.decode(version_output)
            info('The current ROS version is %s.' % ros_version)
            return ros_version
        except (subprocess.CalledProcessError, OSError) as e:
            warning(str(e))
            warning(ros_env_check_notice)
            return 1
        except Exception as e:
            warning(str(e))
            return 1

    def get_roscore_state() -> bool:
        """
        功能描述:检测roscore状态
        参数:无
        """
        max_times = 5
        times = 0
        info('Attempting to get ros master state...')
        cmd = 'import rospy; rospy.get_master().getSystemState()'
        while True:
            try:
                result = subprocess.run(['/usr/bin/env', python_version_for_ros, '-c', cmd], stdout=subprocess.DEVNULL,
                                        stderr=subprocess.DEVNULL)
                if result.returncode == 0:
                    return True
            except Exception as e:
                warning(str(e))
            finally:
                times += 1

            waiting(2, 'get ros master state failed, retry {} times'.format(times))
            if times >= max_times:
                warning('Failed to obtain the ros master status after {} retries. '
                        'Check whether the roscore is running properly.'.format(max_times))
                return False
        return False

    if check_user_os_distro() == 1:
        return 1

    if not is_process_running('roscore'):
        ret = launch_cmd(background_var, ['roscore'], 'roscore', DEFAULT_FLAGS)
        if ret == 1:
            return 1

        if not get_roscore_state():
            return 1

    curr_ros = get_user_ros_distro()
    if curr_ros == 1:
        return 1

    if curr_ros != ros_distro:
        warning('Failed to start MViz due to the incompatible ROS version ' + curr_ros + '.')
        return 1

    return 0


def push_front_to_env(env_var_name, value):
    """
    添加一个value到环境变量中
    参数: env_var_name: 环境变量的名称
    参数: value: 被加到环境变量中的值
    """
    env_var_value = os.getenv(env_var_name)
    os.environ[env_var_name] = '%s:%s' % (value, env_var_value)


def set_env_var(env_var_name, value):
    """
    设置value到环境变量中
    参数: env_var_name: 环境变量的名称
    参数: value: 被设置到环境变量中的值
    """
    env_var_value = os.getenv(env_var_name)
    os.environ[env_var_name] = value


def launch_rviz(background_var):
    """
    功能描述:启动rviz
    参数:background_var 储存运行在后台的进程
    """

    def set_rviz_plugin_env():
        """
        功能描述:设置rviz_plugin的环境变量
        参数:无
        """
        rviz_plugin_path = os.path.join(SCRIPT_PATH, 'rviz_plugin')
        rviz_plugin_lib_path = os.path.join(rviz_plugin_path, 'lib')
        pkgconfig_path = os.path.join(rviz_plugin_lib_path, 'pkgconfig')
        rviz_plugin_share_path = os.path.join(rviz_plugin_path, 'share')
        push_front_to_env('CMAKE_PREFIX_PATH', rviz_plugin_path)
        push_front_to_env('LD_LIBRARY_PATH', rviz_plugin_lib_path)
        push_front_to_env('PKG_CONFIG_PATH', pkgconfig_path)
        push_front_to_env('ROS_PACKAGE_PATH', rviz_plugin_share_path)
        set_env_var('QT_AUTO_SCREEN_SCALE_FACTOR', str(1))  # QT_AUTO_SCREEN_SCALE_FACTOR环境变量表示是否根据屏幕像素密度启用自动缩放，1表示启用

    def set_intel_hardware_decoding_env():
        """
        功能描述:设置Intel硬件解码的环境变量
        参数:无
        """
        media_lib_path = '/opt/media/lib'
        mfx_lib_path = os.path.join(media_lib_path, 'mfx')
        dri_lib_path = os.path.join(media_lib_path, 'dri')
        push_front_to_env('LD_LIBRARY_PATH', media_lib_path)
        push_front_to_env('LD_LIBRARY_PATH', mfx_lib_path)
        push_front_to_env('LD_LIBRARY_PATH', dri_lib_path)

        dri_name = 'iHD'
        os.environ['LIBVA_DRIVER_NAME'] = dri_name

    def set_vehicle_model_env():
        """
        功能描述:设置vehicle_model的环境变量
        参数:无
        """
        vehicle_model_package_path = os.path.join(SCRIPT_PATH, 'vehicle_model')
        push_front_to_env('ROS_PACKAGE_PATH', vehicle_model_package_path)

    def load_robot_models():
        """
        功能描述:加载urdf文件夹下所有的模型配置文件
        参数:无
        """
        for dirpaths, _, urdf_files in os.walk(
                os.path.join(SCRIPT_PATH, 'vehicle_model', 'urdf')):
            for file in urdf_files:
                file_path = os.path.join(dirpaths, file)
                model_description = os.path.splitext(file)[0]
                rosparam_exe = '/opt/ros/' + ros_distro + '/bin/rosparam'
                if not os.path.isfile(rosparam_exe):
                    rosparam_exe = 'rosparam'
                load_param_cmd = [rosparam_exe, 'load', file_path,
                                  model_description]
                load_param_process = subprocess.run(load_param_cmd)
                if load_param_process.returncode == 1:
                    return 1
        return 0

    def is_any_library_not_found(lib_path):
        """
        功能描述:使用ldd检查是否存在依赖库缺失问题
        参数:lib_path-被检查库的路径，例如librviz_plugin.so的路径
        """
        ldd_process = subprocess.run(['/usr/bin/ldd', lib_path],
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
        ldd_out = ldd_process.stdout.decode()
        is_contain_not_found = 0
        for line in ldd_out.splitlines():
            if 'not found' in line:
                warning(line)
                is_contain_not_found = 1

        return is_contain_not_found

    try:
        set_rviz_plugin_env()

        lib_rviz_plugin_path = os.path.join(SCRIPT_PATH, 'rviz_plugin', 'lib',
                                            'librviz_plugin.so')
        if is_any_library_not_found(lib_rviz_plugin_path) == 1:
            warning(ros_env_check_notice)
            return 1

        set_intel_hardware_decoding_env()
        set_vehicle_model_env()

        if load_robot_models() == 1:
            return 1

        flags = DEFAULT_FLAGS
        flags['is_block'] = True
        custom_config_path = os.path.join(SCRIPT_PATH, 'rviz_plugin', 'share',
                                          'rviz_plugin', 'custom.rviz')
        if launch_cmd(background_var,
                      ['rviz', '-d', custom_config_path],
                      'rviz', flags) == 1:
            return 1
        return 0
    except Exception as e:
        warning(str(e))

    return 1


def print_version(cmd):
    """
    功能描述:以特定格式打印出so文件版本或mviz版本
    参数:取得版本号的语句
    """
    print('\033[32m[ INFO] ', end='')
    print(get_version(cmd))


def launch_mviz_server(background_var, bind_ip: str, bind_port: int, is_verify: bool, is_block: bool):
    """
    功能描述:启动mviz
    参数:background_var 储存运行在后台的进程
        bind_ip        mviz server监听的IP
        bind_port      mviz server监听的端口
        is_verify      是否校验内置数据类型的数据范围
        is_block       启动后是运行在后台，还是将脚本阻塞在mviz的这个进程处，在mviz作为转发网关时，就阻塞在此进程处
    """
    flags = DEFAULT_FLAGS
    flags['is_block'] = is_block
    flags['is_output_log'] = True
    print_version("strings " + os.path.dirname(__file__) +
                  "/rviz_plugin" +
                  "/lib" + "/librviz_plugin.so" +
                  " | grep librviz_plugin.so\\ version")
    print_version(os.path.join(SCRIPT_PATH, 'bin', "mviz --version"))

    mviz_path = os.path.join(SCRIPT_PATH, 'bin', 'mviz')
    start_cmd = {
        True: [mviz_path, '-i', bind_ip, '-p', str(bind_port)],
        False: [mviz_path, '-i', bind_ip, '-p', str(bind_port), '-d']
    }

    info('Running MViz and listening on %s:%d' % (bind_ip, bind_port))
    if launch_cmd(background_var, start_cmd[is_verify], 'MViz', flags) == 1:
        return 1
    return 0


def get_version(cmd):
    """
    功能描述:获取so文件版本或mviz版本
    参数:取得版本号的语句
    """
    res = subprocess.Popen(cmd, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out = res.communicate()
    # out[0]是所调用进程传过来的内容，它由进程打印结果和前后缀组成，也就是，b'进程打印结果\n',b'是前缀，\n'是后缀。
    # 此处把内容转化为string，然后截取出进程打印结果
    return str(out[0])[2:len(str(out[0])) - 3]


class InitMvizHostConfig(argparse.Action):
    """
    功能描述: 处理 -i --init参数，用户输入自定义ip和端口
    """

    def __call__(self, parser, namespace, values, option_string=None):
        parser = MvizHostInitializer()
        parser.init_config()


class AllVersion(argparse.Action):
    """
    功能描述:自定义的用于ArgumentParser的action类，用于打印版本号
    参数:取得版本号的语句
    """

    def __call__(self, parser, namespace, values, option_string=None):
        """
        功能描述:打印mviz和so文件的版本号，应接受以下四个参数。
        参数:parser 包含此动作的ArgumentParser对象
            namespace  将由parse_args()返回的Namespace对象。
            values  已关联的命令行参数(命令后面跟的参数而不是命令)，并提供相应的类型转换。
            类型转换由add_argument()的type关键字参数来指定。
            option_string  被用来发起调用此动作的选项字符串。option_string参数是可选的，
            且此参数在动作关联到位置参数时将被略去。
        """
        print(get_version(os.path.join(SCRIPT_PATH, 'bin', "mviz --version")))
        print(get_version("strings " + os.path.dirname(__file__) +
                          "/rviz_plugin" +
                          "/lib" + "/librviz_plugin.so" +
                          " | grep librviz_plugin.so\\ version"))
        sys.exit()


def get_local_ip() -> []:
    """
    功能描述:获取本机IP列表
    参数：无
    """
    cmd = '/bin/hostname -I 2>/dev/null'
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    output_content = process.stdout.readline()
    output_content = str(output_content, encoding='utf-8').strip()

    if len(output_content) == 0:
        return ['127.0.0.1']

    ip_list = output_content.split(' ')
    ip_list.append('127.0.0.1')
    return ip_list


class FileHandler:
    """
    功能描述：文件处理类
    """

    def __init__(self, file_path):
        self.cfg_file = file_path

    def write_file_contents(self, file_content):
        """
        功能描述:写配置文件内容
        参数：无
        """
        file_content_json = json.dumps(file_content, indent=4)
        with open(self.cfg_file, "w") as json_file:
            json_file.write(file_content_json)

    def get_file_contents(self) -> {}:
        """
        功能描述:读配置文件内容
        参数：无
        """
        if not os.path.exists(self.cfg_file):
            return {}
        with open(self.cfg_file, "r") as file_content_json:
            file_content = json.load(file_content_json)
            return file_content


class MvizHostInitializer:
    """
    功能描述：用于处理用户的输入的IP和Port的类
    """

    def __init__(self):
        self.file_handler = FileHandler(os.path.join(SCRIPT_PATH, 'config', 'settings.json'))

    def init_config(self) -> None:
        """
        功能描述:走初始化ip和端口流程 只写入文件
        参数：无
        """
        bind_ip = ''
        bind_port = 0

        print("Initializes the MViz IP address and port configurations.")
        ips = get_local_ip()

        print('Select the IP address that the MDC will connect to: ')
        for ip_idx, ip_str in enumerate(ips):
            print("[%d] %s" % (ip_idx + 1, ip_str))
        idx = input('Enter the number. (If you press Enter, the IP address '
                    'specified by [1] will be used by default.)：\n')
        while True:
            try:
                idx = int(idx) if len(idx) != 0 else 1
                if 1 <= idx <= len(ips):
                    bind_ip = ips[idx - 1]
                    break
                idx = input('Invalid number. Please try again. (If you '
                            'press Enter, the IP address specified by '
                            '[1] will be used by default.)：\n')
            except ValueError:
                idx = input('Invalid number. Please try again. '
                            '(If you press Enter, the IP address specified '
                            'by [1] will be used by default.)：\n')

        port = input('Enter the port that the MDC will connect to. (If you '
                     'press Enter, port 7000 will be used by default.)：\n')
        while True:
            try:
                port = int(port) if len(port) != 0 else 7000
                if 1024 <= port <= 65535:
                    bind_port = port
                    break
                port = input('Invalid port number. Enter a port number '
                             'ranging from 1024 to 65535. '
                             '(If you press Enter, port 7000 will be used '
                             'by default.)：\n')
            except ValueError:
                port = input('Invalid port number. Enter a port number '
                             'ranging from 1024 to 65535. '
                             '(If you press Enter, port 7000 will be used by '
                             'default.)：\n')
        self.init_json_file(bind_ip, bind_port)

    def init_json_file(self, bind_ip, bind_port):
        """
        功能描述:初始化配置文件内容
        参数：用户输入的ip和端口号
        """
        try:
            file_data = self.file_handler.get_file_contents()
        except json.decoder.JSONDecodeError:
            print('An error occurs when parsing the %s file. The file will be reinitialized.'
                  % self.file_handler.cfg_file)
            file_content = {"ip": bind_ip, "port": bind_port, "verification": True}
            self.file_handler.write_file_contents(file_content)
            return
        if "verification" not in file_data:
            file_content = {"ip": bind_ip, "port": bind_port, "verification": True}
        else:
            file_content = {"ip": bind_ip, "port": bind_port,
                            "verification": file_data["verification"]}
        self.file_handler.write_file_contents(file_content)

    def __check_file_contents(self) -> []:
        """
        功能描述:检查配置文件的正确性
        参数：无
        """
        try:
            data = self.file_handler.get_file_contents()
        except json.decoder.JSONDecodeError:
            print('An error occurs when parsing the %s file. Run the ./run.py --init '
                  'command to reinitialize the file.' % self.file_handler.cfg_file)
            sys.exit()

        if "ip" not in data or "port" not in data or "verification" not in data:
            return []

        bind_ip = data["ip"]
        bind_port = data["port"]

        valid_ip_list = get_local_ip()
        if len(valid_ip_list) == 0:
            print('Failed to find any local network devices available.')
            sys.exit()

        if bind_ip not in valid_ip_list:
            print('The IP address %s read from the '
                  '%s file is not the local IP address. '
                  'Run the ./run.py --init command to initialize the local '
                  'network settings of MViz.' % (bind_ip, self.file_handler.cfg_file))
            sys.exit()

        return [bind_ip, bind_port]

    def get_mviz_host_config(self):
        """
        功能描述:读配置文件
        参数：无
        """
        data = self.__check_file_contents()
        if len(data) == 0:
            self.init_config()
            content = self.file_handler.get_file_contents()
            data = [content["ip"], content["port"]]
        return data


class InitVerification(argparse.Action):
    """
    功能描述: 处理 -s --settings参数，用户是否校验数据范围
    """

    def __call__(self, parser, namespace, values, option_string=None):
        parser = VerificationInitializer()
        parser.set_is_check_data_range()


class VerificationInitializer:
    """
    功能描述：用于初始化settings.json文件中verification字段
    """

    def __init__(self):
        self.file_handler = FileHandler(os.path.join(SCRIPT_PATH, 'config', 'settings.json'))

    def init_config(self, file_content):
        """
        功能描述:用户输入决定是否进行数据范围校验
        参数：file_content settings.json文件的内容
        """
        flag = True
        options = ['Y', 'y', 'N', 'n', '']
        while True:
            info('If the verification function is disabled, '
                 'the MViz may exit abnormally when some built-in plug-ins of the RViz receive abnormal data '
                 '(for example, the pose.position field of the Marker class is assigned with an inf value.)')
            input_flag = input('Whether to enable data range verification '
                               '(If yes, press Enter). [Y/n]')
            if input_flag not in options:
                print('Invalid input. Please try again.')
            else:
                if input_flag in ['N', 'n']:
                    flag = False
                break

        file_content["verification"] = flag
        self.file_handler.write_file_contents(file_content)
        print("Set successfully.")

    def set_is_check_data_range(self):
        """
        功能描述:输入-s参数后执行流程的入口
        参数：无
        """
        try:
            result = self.file_handler.get_file_contents()
        except json.decoder.JSONDecodeError:
            print('An error occurs when parsing the %s file. Run the ./run.py --init '
                  'command to reinitialize the file.' % self.file_handler.cfg_file)
            sys.exit()
        self.init_config(result)
        sys.exit()


def check_nvidia_hardware():
    """
    功能描述：在mviz首次启动时检查nvidia硬件是否存在，若存在，则给出提示语句。
    """
    # 这里用来检查MViz是否是第一次启动
    if not os.path.exists(os.path.join(SCRIPT_PATH, 'config',
                                       'settings.json')):
        first_process = subprocess.Popen(['lspci'],
                                         stdout=subprocess.PIPE, shell=False)
        second_process = subprocess.Popen(['grep', '-i', 'nvidia'],
                                          stdin=first_process.stdout,
                                          shell=False,
                                          stdout=subprocess.PIPE)
        result = second_process.wait()
        if result == 0:
            info("The NVIDIA video card is detected in the current "
                 "environment. Currently, "
                 "MViz supports using a NVIDIA video card "
                 "to accelerate H.265 video "
                 "decoding, reducing CPU usage. "
                 "To use this function, check "
                 "whether the video card meets "
                 "the H.265 video decoding requirements "
                 "by referring to section 2 in "
                 "the MDC Application Visualizer "
                 "User Guide and install the video card driver "
                 "of the corresponding "
                 "version (required for hardware decoding).")


def check_python_version():
    """
    检查Python的版本号
    return: 无
    """
    python_version = sys.version.split(' ')
    if len(python_version) > 1:
        python_version = python_version[0]
    else:
        warning("The Python version number cannot be identified.")
        return 1

    big_version, middle_version, _ = [int(x) for x in python_version.split('.')]
    if (big_version, middle_version) < (3, 6):
        warning('The current Python version is %s. Use Python 3.6 or later to '
                'start MViz.' % python_version)
        return 1

    return 0


def main():
    """
    功能描述：mviz 启动脚本的入口函数
    """
    mviz_lib_path = os.path.join(SCRIPT_PATH, 'lib')
    push_front_to_env('LD_LIBRARY_PATH', mviz_lib_path)

    description = "Run MDC_Application_Visualizer: " \
                  "\n   python3 run.py or ./run.py"
    parser = argparse.ArgumentParser(description=description,
                                     allow_abbrev=False)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--version', action=AllVersion, nargs=0,
                       help="Used to query the version number.")
    group.add_argument('-i', '--init', action=InitMvizHostConfig, nargs=0,
                       help='Initializes the MViz IP address and '
                             'port configurations.')
    group.add_argument('-s', '--settings', action=InitVerification, nargs=0,
                       help="Whether to enable data range verification. For details, see the "
                            "\033[3mMDC Application Visualizer User Guide\033[0m.")
    parser.add_argument('-g', '--forward_gateway', action='store_true',
                        help='Only the MViz data communication service is started to '
                             'communicate with the ROS process developed by users.')

    if os.geteuid() == 0:
        warning('The run.py script cannot be executed as root, exit.')
        sys.exit()

    args = parser.parse_args()

    if check_python_version() == 1:
        sys.exit()

    check_nvidia_hardware()
    host_parser = MvizHostInitializer()
    bind_ip, bind_port = host_parser.get_mviz_host_config()
    file_handler = FileHandler(os.path.join(SCRIPT_PATH, 'config', 'settings.json'))
    file_content = file_handler.get_file_contents()
    is_verify = file_content["verification"]

    resource.setrlimit(resource.RLIMIT_CORE, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
    background_var = BackgroundVar([])
    try:
        if check_ros_and_ubuntu_version(background_var) == 1:
            sys.exit()
        if launch_mviz_server(background_var, bind_ip, bind_port, is_verify, args.forward_gateway) == 1:
            sys.exit()
        # 默认情况下，MVIZ不做为转发网关时，会启动rviz
        if not args.forward_gateway:
            if launch_rviz(background_var) == 1:
                sys.exit()
    except KeyboardInterrupt:
        print("\n")
    finally:
        background_var.clean_background()


if __name__ == "__main__":
    main()
